# tinycsv.py - simple csv parser.
# 
# Copyright (c) 2005 M.Shibata <mshibata@emptypage.jp>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to 
# deal in the Software without restriction, including without limitation the 
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
# sell copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
# IN THE SOFTWARE.
# 


_usage = """usage: tinycsv.py [-e encoding] [-t] filename
    -e : specify input file's encoding.
    -t : tab separated format.
"""

CR, LF, CRLF = u'\x0d', u'\x0a', u'\x0d\x0a'
Q = u'"'


class ParseError(Exception):
    """An error occured while parsing the feed."""


class CoreParser:
    def __init__(self, separator=u','):
        self.separator = separator
    
    def feed(self, src):
        if not isinstance(src, unicode):
            src = unicode(src)
        pos = 0
        index_row = 0
        while pos < len(src):
            self.handle_record(index_row)
            self._row = []
            pos = self.parse_record(src, pos)
            self.handle_record_end(index_row)
            index_row += 1
            yield self._row
    
    def parse_record(self, src, pos):
        def parse_and_handle_field(parse_func, handle_func, src, pos, index_col):
            pos2 = parse_func(src, pos)
            field_value = self.decode_field(src[pos:pos2])
            self._row.append(field_value)
            handle_func(index_col, field_value)
            return pos2
        
        index_col = 0
        pos = parse_and_handle_field(self.parse_field, self.handle_field, src, pos, index_col)
        while pos < len(src):
            index_col += 1
            try:
                pos = self.parse_separator(src, pos)
            except ParseError:
                break
            else:
                pos = parse_and_handle_field(self.parse_field, self.handle_field, src, pos, index_col)
        if pos < len(src):
            pos = self.parse_newline(src, pos)
        return pos
    
    def decode_field(self, field):
        return field
    
    def parse_field(self, src, pos):
        while pos < len(src):
            c = src[pos]
            if c.startswith(self.separator) or c in (CR, LF):
                break
            pos += 1
        return pos
    
    def parse_separator(self, src, pos):
        if src[pos:pos + len(self.separator)] == self.separator:
            return pos + len(self.separator)
        raise ParseError
    
    def parse_newline(self, src, pos):
        if src[pos:len(CRLF)] == CRLF:
            return pos + len(CRLF)
        if src[pos] in (CR, LF):
            return pos + len(src[pos])
        raise ParseError
    
    def handle_record(self, index_row):
        return
    
    def handle_record_end(self, index_row):
        return
    
    def handle_field(self, index_col, data):
        return


class CSVParser(CoreParser):
    def __init__(self, separator=','):
        self.separator = separator
    
    def decode_field(self, field):
        if field.startswith(Q):
            return field[1:-1].replace('""', '"')
        else:
            return field
    
    def parse_field(self, src, pos):
        if src[pos] == Q:
            pos += 1
            while pos < len(src):
                if src[pos] == Q:
                    pos += 1
                    if pos == len(src) or src[pos] != Q:
                        return pos
                pos += 1
        else:
            return CoreParser.parse_field(self, src, pos)
        raise ParseError


def main(parser, filename, encoding):
    index_row = 0
    for fields in parser.feed(unicode(file(filename).read(), encoding)):
        print "-- Record %d: --" % index_row
        for index_col in xrange(len(fields)):
            print "%4d: %s" % (index_col, fields[index_col].encode(encoding))
        index_row += 1


if __name__ == '__main__':
    import sys, getopt
    
    try:
        opts, args = getopt.getopt(sys.argv[1:], "te:")
    except getopt.GetoptError:
        print _usage
        sys.exit(2)
    
    separator = ','
    encoding = sys.stdin.encoding
    
    for k, v in opts:
        if k == '-e':
            encoding = v
        if k == '-t':
            separator = '\t'
    
    if not args:
        print _usage
        sys.exit(0)
    
    parser = CSVParser(separator)
    for filename in args:
        main(parser, filename, encoding)

