#!/usr/bin/env python
# encoding: utf-8
# CSV-全銀フォーマット変換フィルタ
# 
# Copyright (c) 2005 M.Shibata <mshibata@emptypage.jp>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to 
# deal in the Software without restriction, including without limitation the 
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
# sell copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
# IN THE SOFTWARE.
# 

import datetime
import kanakana, tinycsv


__usage = {
None: u"""zengin.py - CSV to ZenGin Format filter.

Usage: zengin.py -a account-filename -t template-filename -f format [-d date]
       [-o output-filename] [-b] [csv-filename]

  format: 'kyuuyo' or 'shouyo' or 'furikomi'
  date: YYYY-MM-DD style.
  -b: Tab separated format.
""",

'ja_JP': u"""zengin.py - CSV‐全銀フォーマット変換フィルタ

使い方: zengin.py -a アカウントファイル名 -t テンプレートファイル名 
        -f フォーマット指定 [-d 日付] [-o 出力ファイル名] [-b] [CSV ファイル名]

  フォーマット指定: kyuuyo, shouyo, furikomi のいずれかを指定します。
  日付: YYYY-MM-DD 形式で指定します。
  -b: 元ファイルがタブ区切りのとき指定します。
"""
}


def usage():
    import locale
    langcode, encoding = locale.getdefaultlocale()
    print __usage.get(langcode).encode(encoding or 'ascii', 'replace')


RECORD_HEADER = 1
RECORD_DATA = 2
RECORD_TRAILER = 8
RECORD_END = 9


def load_config(filename):
    cfg = {}
    for line in file(filename):
        line = line.strip()
        if not line or line.startswith('#') or not ':' in line:
            continue
        key, value = [x.strip() for x in line.split(':', 2)]
        cfg[key] = value
    return cfg


def normalize_name(src):
    return kanakana.fullwidth2halfwidth(kanakana.uppercase(kanakana.hiragana2katakana(src)))


def normalize_type(type):
    if isinstance(type, (str, unicode)):
        if u'普' in type:
            return 1
        if u'当' in type:
            return 2
    return int(type)


def normalize_amount(amount):
    if isinstance(amount, (str, unicode)):
        return int(amount.strip('\\').replace(',', ''))
    return int(amount)


class Account:
    def __init__(self, name, bank, branch, type, account, amount=0, ex1=None, ex2=None):
        self.name = normalize_name(name)
        self.bank = int(bank)
        self.branch = int(branch)
        self.type = normalize_type(type)
        self.account = int(account)
        self.ex1, self.ex2 = None, None
        if ex1 is not None:
            self.ex1 = int(ex1)
        if ex2 is not None:
            self.ex2 = int(ex2)
        self.amount = normalize_amount(amount)


class Template(tinycsv.CSVParser):
    
    def __init__(self, separator=','):
        tinycsv.CSVParser.__init__(self, separator)
    
    def set_table(self, table):
        self.table = {}
        i = 0
        for key in table:
            self.table[key] = i
            i += 1
    
    def feed(self, src, first_line_is_header=True):
        param_names = ('name', 'bank', 'branch', 'type', 'account', 'amount')
        linenum = 0
        for fields in tinycsv.CSVParser.feed(self, src):
            linenum += 1
            if first_line_is_header:
                first_line_is_header = False
                continue
            params = [fields[self.table[x]] for x in param_names]
            try:
                yield Account(*params)
            except:
                pass


class Formatter:
    def __init__(self):
        self.code = 0           # 使用コード区分 0:JIS, 1:EBCDIC
        self.company_code = 0
    
    def output(self, my_account, seq_accounts, date):
        self.my_account = my_account
        self.seq_accounts = seq_accounts
        self.date = date
        
        encoding = 'cp932'
        
        yield self.output_header().encode(encoding)
        self.num_accounts = 0
        self.total_amount = 0
        for account in self.seq_accounts:
            self.num_accounts += 1
            self.total_amount += account.amount
            yield self.output_record(account).encode(encoding)
        yield self.output_trailer().encode(encoding)
        yield self.output_footer().encode(encoding)
    
    def output_header(self):
        raise NotImplementedError
    
    def output_record(self, account):
        raise NotImplementedError
    
    def output_trailer(self):
        raise NotImplementedError
    
    def output_footer(self):
        raise NotImplementedError


class KyuuyoFormatter(Formatter):
    def __init__(self):
        Formatter.__init__(self)
        self.type = 11  # 業務種別 11:給与振込（民間）
    
    def output_header(self):
        record = "%1d%2d%1d%010d%-40s%02d%02d%04d%-15s%03d%-15s%1d%07d" % (
            RECORD_HEADER,            # レコードの種類
            self.type,                # 業務種別
            self.code,                # 使用コード区分 0:JIS, 1:EBCDIC
            self.company_code,        # 銀行が採番した会社コード
            self.my_account.name,     # 会社名
            self.date.month, self.date.day,   # 振込指定日
            self.my_account.bank,     # 取引銀行番号（統一金融機関番号）
            '',                       # 銀行名（任意）
            self.my_account.branch,   # 支店番号
            '',                       # 支店名（任意）
            self.my_account.type,     # 預金種目 1:普通, 2:当座（任意）
            self.my_account.account,  # 口座番号
        )
        return "%-120s" % record
    
    def output_record(self, account):
        record = "%1d%04d%-15s%03d%-15s%04d%1d%07d%-30s%010d%1d" % (
            RECORD_DATA,        # レコードの種類 2: データ・レコード
            account.bank,       # 振込先銀行番号
            '',                 # 振込先銀行名 ** NOT IMPLEMENTED **
            account.branch,     # 振込先支店番号
            '',                 # 振込先支店名 ** NOT IMPLEMENTED **
            0,                  # 手形交換所番号（任意）
            account.type,       # 預金種目 1:普通, 2:当座
            account.account,    # 今座番号
            account.name,       # 振込先預金者名
            account.amount,    # 振込金額
            0,                  # 新規コード 1:第１回振込分, 2:変更分, 0:その他
        )
        if account.ex1 is not None:
            record += "%010d" % account.ex1     # 社員番号（任意）
        if account.ex2 is not None:
            record += "%010d" % account.ex2     # 所属コード（任意）
        return "%-120s" % record
    
    def output_trailer(self):
        return "%-120s" % ("%1d%06d%012d" % (RECORD_TRAILER, self.num_accounts, self.total_amount))
    
    def output_footer(self):
        return "%-120s" % ("%1d" % RECORD_END)


class ShouyoFormatter(KyuuyoFormatter):
    def __init__(self,):
        Formatter.__init__(self)
        self.type = 12  # 業務種別 12:賞与振込（民間）


class FurikomiFormatter(KyuuyoFormatter):
    def __init__(self,):
        Formatter.__init__(self)
        self.type = 21  # 業務種別 21:総合振込


if __name__ == '__main__':
    import sys, getopt
    
    try:
        opts, args = getopt.getopt(sys.argv[1:], "t:a:f:d:o:b")
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    
    my_account, template, formatter_class = None, None, None
    first_line_is_header = True
    encoding = sys.stdin.encoding
    date = datetime.date.today()
    fp_in = sys.stdin
    fp_out = sys.stdout
    separator = ','
    fields = ''
    
    for k, v in opts:
        if k == '-t':
            cfg = load_config(v)
            if cfg.has_key('firstline'):
                first_line_is_header = (cfg['firstline'] == 'header')
            if cfg.has_key('encoding'):
                encoding  = cfg['encoding']
            fields = cfg['fields']
        if k == '-a':
            cfg = load_config(v)
            param_names = ('name', 'bank', 'branch', 'type', 'account')
            params = [unicode(cfg[x], cfg['encoding']) for x in param_names]
            my_account = Account(*params)
        if k == '-f':
            formatter_class = {'kyuuyo': KyuuyoFormatter,
                               'kyuyo': KyuuyoFormatter,
                               'shouyo': ShouyoFormatter,
                               'furikomi': FurikomiFormatter}[v]
        if k == '-d':
            date = datetime.date(*[int(x) for x in v.split('-')])
        if k == '-o':
            fp_out = file(v, 'wb')
        if k == '-b':
            separator = '\t'
    
    if args:
        fp_in = file(args[0])
    
    template = Template(separator)
    template.set_table([x.strip() for x in fields.split(',')])
    if my_account and template and formatter_class:
        src = unicode(fp_in.read(), encoding, 'ignore')
        formatter = formatter_class()
        for line in formatter.output(my_account, template.feed(src, first_line_is_header), date):
            fp_out.write(line + '\n')
        sys.exit()
    
    usage()

